@@ -24,7 +24,9 @@ libtoolize --force --copy --install || glibtoolize --force --copy --install
 
 # makes a aclocal.m4 which includes the automake macros for autconf
 
-aclocal &&
+mkdir -p m4 &&
+
+aclocal -I m4 &&
 
 # generate configure script
 
