@@ -32,6 +32,16 @@
 #include <setjmp.h>
 #ifdef HAVE_MCHECK
 # include <mcheck.h>
+# include <pthread.h>
+extern void mcheck_init(int flag);
+extern void* (*malloc_l)(size_t size);
+extern void (*free_l)(void* addr);
+extern void* (*realloc_l)(void* addr, size_t size);
+#else
+# define malloc_l(size) malloc(size)
+# define free_l(addr) free(addr)
+# define realloc_l(addr, size) realloc(addr, size)
+# define mcheck_init(flag)
 #endif
 
 #if !defined(FORCE_LL) && !defined(BUGGY_LONG_LONG)
@@ -139,8 +149,8 @@ typedef unsigned OCTABYTE_TYPE UOctabyte;
 #define F_FALSE (FLAG(0!=0))
 
 typedef struct {
-  Cell* sp;
-  Float* fp;
+  Cell* spx;
+  Float* fpx;
 } ptrpair;
 
 /* define this false if you want native division */
@@ -223,6 +233,16 @@ typedef DOUBLE_UCELL_TYPE UDCell;
 
 #endif /* !defined(BUGGY_LONG_LONG) */
 
+#define DIV_SETUP(d1, n1, n2) \
+DCell d1; \
+Cell sign = n2 < 0; \
+if(sign) { \
+  n2 = -n2; \
+  D_IS(d1, (n1<0 ? 0: n2-1), -n1); \
+} else { \
+  D_IS(d1, (n1<0 ? n2-1 : 0), n1); \
+}
+
 typedef union {
   struct {
 #if defined(WORDS_BIGENDIAN)||defined(BUGGY_LONG_LONG)
@@ -421,7 +441,7 @@ Label *gforth_engine2(Xt *ip sr_proto);
 Label *gforth_engine3(Xt *ip sr_proto);
 
 Cell gforth_main(int argc, char **argv, char **env);
-void gforth_args(int argc, char ** argv, char ** path, char ** imagename);
+int gforth_args(int argc, char ** argv, char ** path, char ** imagename);
 ImageHeader* gforth_loader(char* imagename, char* path);
 user_area* gforth_stacks(Cell dsize, Cell rsize, Cell fsize, Cell lsize);
 void gforth_free_stacks(user_area* t);
